'' =================================================================================================
''
''   File....... jm_lcd_pcf8574_demo.spin
''   Purpose.... Demonstrates jm_lcd_pcf8574 interface
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (C) 2018-2023 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon.mcphalen@gmail.com
''   Started....
''   Updated.... 24 JAN 2022
''
'' =================================================================================================


con { timing }

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000                                          ' use 5MHz crystal

  CLK_FREQ = (_clkmode >> 6) * _xinfreq                         ' system freq as a constant
  MS_001   = CLK_FREQ / 1_000                                   ' ticks in 1ms
  US_001   = CLK_FREQ / 1_000_000                               ' ticks in 1us


con { terminal }

  BR_TERM  = 115_200                                            ' terminal baud rate


con { fixed io pins }

  RX1      = 31  { I }                                          ' serial / programming
  TX1      = 30  { O }

  SDA1     = 29  { I/O }                                        ' i2c / eeprom
  SCL1     = 28  { I/O }


con { app io pins }

  SDAX     = 29  { I/O }                                        ' I2C bus
  SCLX     = 28  { I/O }


con

  #true,  ON, OFF
  #false, NO, YES


  LCD_ADDR  = %111                                              ' LCD I2C address

  LCD_WIDTH =   16                                              ' LCD size (16 or 20)
  LCD_LINES =    2                                              ' 2 or 4


obj

' main                                                          ' * master Spin cog
  time : "jm_time_80"                                           '   timing and delays (80MHz
  lcd  : "jm_lcd_pcf8574"                                       '   pcf8574-based LCD (I2C)
  term : "jm_fullduplexserial"                                  ' * serial IO for terminal

' * uses cog when loaded


dat

  Mouth0        byte      $0E, $1F, $1C, $18, $1C, $1F, $0E, $00        ' custom characters
  Mouth1        byte      $0E, $1F, $1F, $18, $1F, $1F, $0E, $00
  Mouth2        byte      $0E, $1F, $1F, $1F, $1F, $1F, $0E, $00
  Smile         byte      $00, $0A, $0A, $00, $11, $0E, $06, $00

  Banner1       byte      $20, $20, $20, $20, $20, $20, $20, $20
                byte      $20, $20, $20, $20, $20, $20, $20, $20
                byte      $20, $20, $20, $20, $4A, $4F, $4E, $4E
                byte      $59, $4D, $41, $43, $20, $52, $55, $4C
                byte      $45, $53, $21, $03, $20, $20, $20, $20
                byte      $00

  Banner2       byte      "P1 I2C LCD DEMO     ", 0


pub main

  setup

  lcd.cmd(lcd#CLS)
  time.pause(250)

  repeat
    lcd.cmd(lcd#CLS)
    time.pause(250)

    scroll_animate_demo
    time.pause(2000)

    numbers_demo
    time.pause(500)

    backlight_demo


pub scroll_animate_demo | pos, char, frame, newchar

  ' scroll on 1st line

  repeat pos from 0 to LCD_WIDTH                                ' loop through scrolling string
    lcd.crsr_xy(0, 0)                                           ' move to start of top line
    if (LCD_WIDTH == 20)
      lcd.sub_str(@Banner1, pos, LCD_WIDTH)                     ' move part of string to display
    else
      lcd.sub_str(@Banner1+4, pos, LCD_WIDTH)
    time.pause(125)                                             ' control scrolling speed

  ' animate 2nd line

  repeat pos from 0 to LCD_WIDTH-1                              ' scroll through all chars
    char := byte[@Banner2][pos]                                 ' get char from banner2
    repeat frame from 1 to 6                                    ' loop through animation frames
      lcd.crsr_xy(pos, 1)                                       ' position cursor
      newchar := lookup(frame : 0, 1, 2, 2, 1, char)            ' get char for frame
      lcd.out(newchar)                                          ' write it
      time.pause(100)                                           ' short, inter-frame delay


pub numbers_demo | tf, tc

  lcd.cmd(lcd#CLS)
  lcd.str(string("TEMP"))

  repeat tf from 98_6 to 105_7                                  ' fahrenheit range
    tc := (tf - 32_0) * 5 / 9                                   ' convert to celsius
    lcd.crsr_xy(6, 0)
    lcd.jdpdec(tf, 1, 5, " ")                                   ' justified decimal point, 1 digit
    lcd.out("F")
    lcd.crsr_xy(6, 1)
    lcd.jdpdec(tc, 1, 5, " ")
    lcd.out("C")
    time.pause(150)


pub backlight_demo

  repeat 3                                                      ' backlight control
    lcd.backlight(false)
    time.pause(500)
    lcd.backlight(true)
    time.pause(500)


con

  CC_MOUTH0 = 0
  CC_MOUTH1 = 1
  CC_MOUTH2 = 2
  CC_SMILE  = 3


pub setup

  time.start

  lcd.start(LCD_ADDR, SCLX, SDAX)                               ' connect to I2C @ 100kHz w/o pullups
  lcd.set_char(CC_MOUTH0, @Mouth0)                              ' define custom characters
  lcd.set_char(CC_MOUTH1, @Mouth1)
  lcd.set_char(CC_MOUTH2, @Mouth2)
  lcd.set_char(CC_SMILE , @Smile)

  ' if 20-char display, move smiley in Banner1

  if (LCD_WIDTH == 20)
    Banner1[35] := $20
    Banner1[36] := CC_SMILE

  term.tstart(BR_TERM)                                          ' start terminal io


pub wait_for_terminal(clear, delay)

  ' Download to RAM with F10
  ' F12 to open PST
  ' Click [Enable]
  ' Press Enter

  term.rxflush
  term.rx
  if (clear)
    term.tx(term#CLS)
    time.pause(delay #> 0)


con { license }

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

}}